! DESCRIPTION: Parallel code for approximating the area under a curve using a 
!                  left Riemann sum.
! AUTHOR:      Aaron Weeden, Shodor Education Foundation, Inc.
! DATE:        September 2011
!              Revised December 2011
! EXAMPLE USAGE:
!        To run a program (e.g. serial) with the default domain width and 
!          default number of rectangles, use:    ./area.serial
!        To run a program (e.g. openmp) with the default domain width and 
!          100000 rectangles, use:    ./area.openmp -n 100000
!        To run a program (e.g. mpi) with a domain from 100.0 to 200.0 and the
!          default number of rectangles, use:   ./area.mpi -l 100.0 -r 200.0
!        To run a program (e.g. hybrid) with a domain from 50.1 to 75.5 and
!          with 50000 rectangles, use:   ./area.hybrid -l 50.1 -r 75.5 -n 50000
!        To run a program (e.g. serial) with the default left x-boundary of the
!          domain and 500.0 as the right x-boundary of the domain with the
!          default number of rectangles, use:  ./area.serial -r 500.0

PROGRAM area
IMPLICIT NONE

!!!!!!!!!!!!!!!!!!!!!!!
! Libraries to import !
!!!!!!!!!!!!!!!!!!!!!!!
#ifdef MPI
    INCLUDE "mpif.h"
#endif

#ifdef OPENMP
    INTEGER, EXTERNAL :: OMP_GET_THREAD_NUM, OMP_GET_NUM_THREADS
#endif

! Declare the constants, variables, and data structures.
INTEGER :: OUR_RANK = 0, OUR_NUMBER_OF_THREADS = 1, &
    OUR_NUMBER_OF_RECTANGLES = 0, NUMBER_OF_RECTANGLES = 10, &
    NUMBER_OF_PROCESSES = 1, my_thread_num = 0, &
    my_current_rectangle_id = 0, ierror = 0
DOUBLE PRECISION :: OUR_X_LEFT = 0.0, X_LEFT = 0.0, X_RIGHT = 10.0, &
    WIDTH = 0.0, RECTANGLE_WIDTH = 0.0, our_total_sum = 0.0, &
    the_total_sum = 0.0, my_current_rectangle_left = 0.0, &
    my_current_rectangle_height = 0.0
DOUBLE PRECISION, ALLOCATABLE, DIMENSION(:) :: our_areas

! For command line parsing
INTEGER :: i, print_usage = 0
CHARACTER(len=32) :: arg

#ifdef MPI
    ! Initialize the MPI environment
    CALL MPI_INIT(ierror)
    CALL MPI_COMM_RANK(MPI_COMM_WORLD, OUR_RANK, ierror)
    CALL MPI_COMM_SIZE(MPI_COMM_WORLD, NUMBER_OF_PROCESSES, ierror)
#endif

! Parse command line arguments
i = 1
DO WHILE(i <= COMMAND_ARGUMENT_COUNT())
    CALL GET_COMMAND_ARGUMENT(i, arg)
    SELECT CASE (arg)
    CASE ('-n')
        IF(i == COMMAND_ARGUMENT_COUNT()) THEN
            print_usage = 1
        END IF
        CALL GET_COMMAND_ARGUMENT(i+1, arg)
        READ(arg, *) NUMBER_OF_RECTANGLES
        i = i + 1
    CASE ('-l')
        IF(i == COMMAND_ARGUMENT_COUNT()) THEN
            print_usage = 1
        END IF
        CALL GET_COMMAND_ARGUMENT(i+1, arg)
        READ(arg, *) X_LEFT
        i = i + 1
    CASE ('-r')
        IF(i == COMMAND_ARGUMENT_COUNT()) THEN
            print_usage = 1
        END IF
        CALL GET_COMMAND_ARGUMENT(i+1, arg)
        READ(arg, *) X_RIGHT
        i = i + 1
    CASE DEFAULT
        print_usage = 1
    END SELECT
    i = i + 1
END DO
IF(print_usage == 1) THEN
    PRINT *, "Available command line arguments:"
    PRINT *, "\t-n NUMBER_OF_RECTANGLES\n\t-l X_LEFT\n\t-r X_RIGHT"
    STOP
END IF

! Allocate the array.
ALLOCATE(our_areas(NUMBER_OF_RECTANGLES))

#ifdef OPENMP
!$OMP parallel private(my_thread_num)
    my_thread_num = OMP_GET_THREAD_NUM()
#endif

    IF(my_thread_num == 0) THEN
        ! Calculate the overall width of the domain of the
        !  function and the width of a rectangle.
        WIDTH = X_RIGHT - X_LEFT;
        RECTANGLE_WIDTH = WIDTH / NUMBER_OF_RECTANGLES
    END IF
    IF(my_thread_num == MODULO(1, OUR_NUMBER_OF_THREADS)) THEN
        ! Calculate the number of rectangles for 
        !  which the process is responsible.
        OUR_NUMBER_OF_RECTANGLES = NUMBER_OF_RECTANGLES / NUMBER_OF_PROCESSES
        IF(OUR_RANK == NUMBER_OF_PROCESSES - 1) THEN
            OUR_NUMBER_OF_RECTANGLES = OUR_NUMBER_OF_RECTANGLES + &
                MODULO(NUMBER_OF_RECTANGLES, NUMBER_OF_PROCESSES)
        END IF
    END IF
    IF(my_thread_num == MODULO(2, OUR_NUMBER_OF_THREADS)) THEN
        ! Calculate the left x-boundary of the process.
        OUR_X_LEFT = (OUR_RANK * (NUMBER_OF_RECTANGLES / NUMBER_OF_PROCESSES)) &
                        + X_LEFT
    END IF

#ifdef OPENMP
    !$OMP end parallel
#endif

#ifdef OPENMP
    !$OMP parallel do private (my_current_rectangle_id, my_current_rectangle_left, my_current_rectangle_height)
#endif
    DO my_current_rectangle_id = 0, OUR_NUMBER_OF_RECTANGLES - 1, 1
    ! Calculate the x-value of the left side of the rectangle
        my_current_rectangle_left = OUR_X_LEFT + my_current_rectangle_id &
        * RECTANGLE_WIDTH

            ! Calculate the height of the rectangle
        my_current_rectangle_height = FUNC(my_current_rectangle_left)

        ! Calculate the area of the rectangle
        our_areas(my_current_rectangle_id + 1) = RECTANGLE_WIDTH &
        * my_current_rectangle_height
    END DO
#ifdef OPENMP
    !$OMP end parallel do
#endif

! Calculate the total sum for the process
DO my_current_rectangle_id = 1, OUR_NUMBER_OF_RECTANGLES, 1
    our_total_sum = our_total_sum + our_areas(my_current_rectangle_id)
END DO

#ifdef MPI
    ! Calculate the overall total sum
    CALL MPI_REDUCE(our_total_sum, the_total_sum, 1, MPI_DOUBLE_PRECISION, &
        MPI_SUM, 0, MPI_COMM_WORLD, ierror)
#else
    the_total_sum = our_total_sum
#endif

#ifdef SHOW_RESULTS
    ! Print the total sum
    IF(OUR_RANK == 0) THEN
        PRINT *, the_total_sum
    END IF
#endif

! Deallocate the array
DEALLOCATE(our_areas)

#ifdef MPI
    ! Finalize the MPI environment
    CALL MPI_FINALIZE(ierror)
#endif

CONTAINS
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Mathematical function under whose curve we wish to approximate the area !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
DOUBLE PRECISION FUNCTION FUNC(x)
    IMPLICIT NONE

    DOUBLE PRECISION, INTENT(IN) :: x
    FUNC = x * x
END FUNCTION FUNC

END PROGRAM area
